@@ -18,7 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
     Evan Klinger - eklinger@phash.org
-    David Starkweather - dstarkweather@phash.org
+    D Grant Starkweather - dstarkweather@phash.org
 
 */
 
@@ -41,6 +41,10 @@
 #include <stdint.h>
 #include "CImg.h"
 
+#ifdef HAVE_PTHREAD
+#include <pthread.h>
+#endif
+
 using namespace cimg_library;
 using namespace std;
 
@@ -60,25 +64,34 @@ typedef unsigned long long ulong64;
 typedef signed long long long64;
 #endif
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 const int MaxFileSize = (1<<30); /* 1GB file size limit (for mvp files) */
 const off_t HeaderSize = 64;     /* header size for mvp file */
 
-
-const char *mvptag = "pHashMVPfile2009";
+const char mvptag[] = "pHashMVPfile2009";
 
 typedef enum ph_mvp_retcode {
     PH_SUCCESS = 0,   /* success */
     PH_ERRPGSIZE,     /* page size error */
-    PH_ERRFILE,       /* file operations */
+    PH_ERRSMPGSIZE,
+    PH_ERRFILEOPEN,       /* file open err */
+    PH_ERRFILETRUNC,
+    PH_ERRFILESEEK,
     PH_ERRMMAP,        /* mmap'ing error */
     PH_ERRMSYNC,       /* msync error */
     PH_ERRTRUNC,       /* error truncating file */
     PH_ERRSAVEMVP,      /* could not save mvp file */
-    PH_ERRARG,   /* null arg */
-    PH_ERRMEM,       /* mem alloc error - not enough available memory */
+    PH_ERRARG,   /* general arg err*/
+    PH_ERRNULLARG,   /* null arg */
+    PH_ERRMEM,       /* general memory error  */
+    PH_ERRMEMALLOC,  /* mem alloc error */
     PH_ERRNTYPE,      /* unrecognized node type */
     PH_ERRCAP,     /* more results found than can be supported in ret array */
     PH_ERRFILETYPE,  /*unrecognized file type  */
+    PH_ERRDIST,      /*not enough variance in distances */
 }MVPRetCode;
 
 
@@ -100,18 +113,25 @@ typedef struct ph_datapoint {
     char *id;
     void *hash;
     float *path;
-    uint16_t hash_length;
+    uint32_t hash_length;
     uint8_t hash_type;
 }DP;
 
+typedef struct ph_slice {
+    DP **hash_p;
+    int n;
+    void *hash_params;
+} slice;
+
 /* call back function for mvp tree functions - to performa distance calc.'s*/
 typedef float (*hash_compareCB)(DP *pointA, DP *pointB);
 
 typedef struct ph_mvp_file {
-    const char *filename;   /* name of db to use */
+    char *filename;   /* name of db to use */
     char *buf;
     off_t file_pos;
     int fd;
+    uint8_t filenumber;
     uint8_t nbdbfiles;
     uint8_t branchfactor; /*branch factor of tree, M(=2)*/
 
@@ -135,14 +155,7 @@ typedef struct ph_mvp_file {
 
 
 /* convenience function to set var's of mvp tree */
-void ph_mvp_init(MVPFile *m){
-    m->branchfactor = 2;
-    m->pathlength = 5;
-    m->leafcapacity = 25;
-    m->pgsize = sysconf(_SC_PAGE_SIZE);     /* use host page size */
-    return;
-}
-
+void ph_mvp_init(MVPFile *m);
 
 /*! /brief Radon Projection info
  */
@@ -186,10 +199,14 @@ typedef struct ph_match{
     uint32_t length;    /*length of match between 2 files */
 } TxtMatch;
 
+#ifdef HAVE_PTHREAD
+int ph_num_threads();
+#endif
+
 /* /brief alloc a single data point
  *  allocates path array, does nto set id or path
  */
-DP* ph_malloc_datapoint(int hashtype, int pathlength);
+ DP* ph_malloc_datapoint(int hashtype);
 
 /** /brief free a datapoint and its path
  *
@@ -247,7 +264,7 @@ int ph_crosscorr(const Digest &x,const Digest &y,double &pcc, double threshold =
  *  /param N      - int value for the number of angles to consider. 
  *  /return       - less than 0 for error
  */
-int ph_image_digest(const CImg<uint8_t> &img,double sigma, double gamma,Digest &digest,int N=180);
+int _ph_image_digest(const CImg<uint8_t> &img,double sigma, double gamma,Digest &digest,int N=180);
 
 /*! /brief image digest
  *  Compute the image digest given the file name.
@@ -270,7 +287,7 @@ int ph_image_digest(const char *file, double sigma, double gamma, Digest &digest
  *  /param theshold - double value for the threshold
  *  /return int 0 (false) for different images, 1 (true) for same image, less than 0 for error
  */
-int ph_compare_images(const CImg<uint8_t> &imA,const CImg<uint8_t> &imB,double &pcc, double sigma = 3.5, double gamma = 1.0,int N=180,double threshold=0.90);
+int _ph_compare_images(const CImg<uint8_t> &imA,const CImg<uint8_t> &imB,double &pcc, double sigma = 3.5, double gamma = 1.0,int N=180,double threshold=0.90);
 
 /*! /brief compare 2 images
  *  Compare 2 images given the file names
@@ -289,7 +306,7 @@ int ph_compare_images(const char *file1, const char *file2,double &pcc, double s
  *  /param N - int denoting the size of the square matrix to create.
  *  /return CImg<double> size NxN containing the dct matrix
  */
-CImg<float>* ph_dct_matrix(const int N);
+static CImg<float>* ph_dct_matrix(const int N);
 
 /*! /brief compute dct robust image hash
  *  /param file string variable for name of file
@@ -298,10 +315,14 @@ CImg<float>* ph_dct_matrix(const int N);
  */
 int ph_dct_imagehash(const char* file,ulong64 &hash);
 
+#ifdef HAVE_PTHREAD
+DP** ph_dct_image_hashes(char *files[], int count, int threads = 0);
+#endif
+static CImgList<uint8_t>* ph_getKeyFramesFromVideo(const char *filename);
 
-CImgList<uint8_t>* ph_getKeyFramesFromVideo(const char *filename);
+ulong64* ph_dct_videohash(const char *filename, int &Length);
 
-ulong64* ph_dct_videohash(const char *filename, int &Lenght);
+DP** ph_dct_video_hashes(char *files[], int count, int threads = 0);
 
 double ph_dct_videohash_dist(ulong64 *hashA, int N1, ulong64 *hashB, int N2, int threshold=21);
 
@@ -322,12 +343,37 @@ int ph_hamming_distance(const ulong64 hash1,const ulong64 hash2);
 
 DP** ph_read_imagehashes(const char *dirname,int capacity, int &count);
 
+/** /brief create MH image hash for filename image
+*   /param filename - string name of image file
+*   /param N - (out) int value for length of image hash returned
+*   /param alpha - int scale factor for marr wavelet (default=2)
+*   /param lvl   - int level of scale factor (default = 1)
+*   /return uint8_t array
+**/
+uint8_t* ph_mh_imagehash(const char *filename, int &N, float alpha=2.0f, float lvl = 1.0f);
+
+/** /brief count number bits set in given byte
+*   /param val - uint8_t byte value
+*   /return int value for number of bits set
+**/
+int ph_bitcount8(uint8_t val);
+
+/** /brief compute hamming distance between two byte arrays
+ *  /param hashA - byte array for first hash
+ *  /param lenA - int length of hashA 
+ *  /param hashB - byte array for second hash
+ *  /param lenB - int length of hashB
+ *  /return double value for normalized hamming distance
+ **/
+double ph_hammingdistance2(uint8_t *hashA, int lenA, uint8_t *hashB, int lenB);
+
 /** /brief get all the filenames in specified directory
  *  /param dirname - string value for path and filename
  *  /param cap - int value for upper limit to number of files
  *  /param count - int value for number of file names returned
  *  /return array of pointers to string file names (NULL for error)
  **/
+
 char** ph_readfilenames(const char *dirname,int &count);
 
 /** /brief read a datapoint (aux function)
@@ -356,14 +402,14 @@ off_t ph_save_datapoint(DP *new_dp, MVPFile *m);
  *  /param m - MVPFile
  *  /return MVPFile - ptr to new struct containing the mmap info
  **/
-MVPFile* _ph_map_mvpfile(uint8_t filenumber, off_t offset, MVPFile *m);
+MVPRetCode _ph_map_mvpfile(uint8_t filenumber, off_t offset, MVPFile *m, MVPFile *m2);
 
 /** /brief unmap/map from m2 to m
  *  /param filenumber - uint8_t filenumber of m2
  *  /param orig_pos   = off_t offset into original file in m.
  *  /return void
  **/
-void _ph_unmap_mvpfile(uint8_t filenumber, off_t orig_pos, MVPFile *m, MVPFile *m2);
+MVPRetCode _ph_unmap_mvpfile(uint8_t filenumber, off_t orig_pos, MVPFile *m, MVPFile *m2);
 
 /**
  * callback function for dct image hash use in mvptree structure.
@@ -376,12 +422,12 @@ float hammingdistance(DP *pntA, DP *pntB);
  *  /param knearest - int capacity of results array.
  *  /param radius - float value of radius of values to consider
  *  /param results - DP array of points of result
- *  /param count - int* number of results found (out)
+ *  /param nbfound - int number of results found (out)
  *  /param level - int value to track recursion depth.
  *  /return MVPRetCode
 **/
-MVPRetCode ph_query_mvptree(MVPFile *m, DP *query, int knearest, float radius,
-			    DP **results, int *count, int level);
+static MVPRetCode _ph_query_mvptree(MVPFile *m, DP *query, int knearest, float radius, 
+                float threshold, DP **results, int &nbfound, int level);
 
 /**  /brief query mvptree function
  *   /param m - MVPFile file state info
@@ -389,10 +435,10 @@ MVPRetCode ph_query_mvptree(MVPFile *m, DP *query, int knearest, float radius,
  *   /param knearest - int capacity of results array
  *   /param radius  - float radius to consider in query
  *   /param results - DP** list of pointers to results found
- *   /param count -  int number of results found (out)
+ *   /param nbfound -  int number of results found (out)
  **/
 MVPRetCode ph_query_mvptree(MVPFile *m, DP *query, int knearest, float radius,
-			    DP **results, int *count);
+		float threshold,   DP **results, int &nbfound);
 
 /** /brief save dp points to a file (aux func)
  *  /param m - MVPFile state information of file
@@ -402,7 +448,7 @@ MVPRetCode ph_query_mvptree(MVPFile *m, DP *query, int knearest, float radius,
  *  /param level - int track recursion level
  *  /return FileIndex* - fileno and offset into file.
 **/
-FileIndex* ph_save_mvptree(MVPFile *m, DP **points, int nbpoints, int saveall_flag, int level);
+static MVPRetCode _ph_save_mvptree(MVPFile *m, DP **points, int nbpoints, int saveall_flag, int level,FileIndex *pOffset);
 
 /** /brief save points to mvp file 
  *  /param m - MVPFile state info of file
@@ -419,7 +465,7 @@ MVPRetCode ph_save_mvptree(MVPFile *m, DP **points, int nbpoints);
  *   /param level - int track recursion level
  *   /return MVPRetCode
  **/
-MVPRetCode ph_add_mvptree(MVPFile *m, DP *new_dp, int level);
+static MVPRetCode _ph_add_mvptree(MVPFile *m, DP *new_dp, int level);
 
 /** /brief add a list of points to mvp file
     /param m - MVPFile state information of file.
@@ -427,7 +473,7 @@ MVPRetCode ph_add_mvptree(MVPFile *m, DP *new_dp, int level);
     /param nbpoints - int number of points
     /return int - number of points added, neg for error
 **/
-int ph_add_mvptree(MVPFile *m, DP **points, int nbpoints);
+MVPRetCode ph_add_mvptree(MVPFile *m, DP **points, int nbpoints, int &nbsaved);
 
 /** /brief textual hash for file
  *  /param filename - char* name of file
@@ -448,7 +494,7 @@ TxtMatch* ph_compare_text_hashes(TxtHashPoint *hash1, int N1, TxtHashPoint *hash
 
 /* random char mapping for textual hash */
 
-ulong64 textkeys[256] = {
+static const ulong64 textkeys[256] = {
     15498727785010036736LLU,
     7275080914684608512LLU,
     14445630958268841984LLU,
@@ -707,5 +753,8 @@ ulong64 textkeys[256] = {
     11569021017716162560LLU
 };
 
+#ifdef __cplusplus
+}
+#endif
 
 #endif
